#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: ec2_eni
version_added: 1.0.0
short_description: Create and optionally attach an Elastic Network Interface (ENI) to an instance
description:
  - Create and optionally attach an Elastic Network Interface (ENI) to an instance.
  - If O(eni_id) or O(private_ip_address) is provided, the existing ENI (if any) will be modified.
  - The O(attached) parameter controls the attachment status of the network interface.
author:
  - "Rob White (@wimnat)"
  - "Mike Healey (@healem)"
options:
  eni_id:
    description:
      - The ID of the ENI (to modify).
      - If O(eni_id=None) and O(state=present), a new ENI will be created.
    type: str
  instance_id:
    description:
      - Instance ID that you wish to attach ENI to.
    type: str
  private_ip_address:
    description:
      - Private IP address.
    type: str
  subnet_id:
    description:
      - ID of subnet in which to create the ENI.
    type: str
  description:
    description:
      - Optional description of the ENI.
    type: str
  security_groups:
    description:
      - List of security groups associated with the interface.
      - Ignored when O(state=absent).
    type: list
    elements: str
    default: []
  state:
    description:
      - Create or delete ENI.
    default: present
    choices: [ 'present', 'absent' ]
    type: str
  device_index:
    description:
      - The index of the device for the network interface attachment on the instance.
    default: 0
    type: int
  attached:
    description:
      - Specifies if network interface should be attached or detached from instance. If omitted, attachment status
        won't change
    type: bool
  force_detach:
    description:
      - Force detachment of the interface. This applies either when explicitly detaching the interface by setting O(instance_id=None)
        or when deleting an interface with O(state=absent).
    default: false
    type: bool
  delete_on_termination:
    description:
      - Delete the interface when the instance it is attached to is terminated. You can only specify this flag when the
        interface is being modified, not on creation.
    required: false
    type: bool
  source_dest_check:
    description:
      - By default, interfaces perform source/destination checks. NAT instances however need this check to be disabled.
        You can only specify this flag when the interface is being modified, not on creation.
    required: false
    type: bool
  secondary_private_ip_addresses:
    description:
      - A list of IP addresses to assign as secondary IP addresses to the network interface.
      - This option is mutually exclusive of O(secondary_private_ip_address_count).
    required: false
    type: list
    elements: str
  purge_secondary_private_ip_addresses:
    description:
      - To be used with O(secondary_private_ip_addresses) to determine whether or not to remove any secondary IP addresses other than those specified.
      - Set O(secondary_private_ip_addresses=[]) to purge all secondary addresses.
    default: false
    type: bool
  secondary_private_ip_address_count:
    description:
      - The number of secondary IP addresses to assign to the network interface.
      - This option is mutually exclusive of O(secondary_private_ip_addresses).
    required: false
    type: int
  allow_reassignment:
    description:
      - Indicates whether to allow an IP address that is already assigned to another network interface or instance
        to be reassigned to the specified network interface.
    required: false
    default: false
    type: bool
  name:
    description:
      - Name for the ENI. This will create a tag with the key V(Name) and the value assigned here.
      - This can be used in conjunction with O(subnet_id) as another means of identifiying a network interface.
      - AWS does not enforce unique V(Name) tags, so duplicate names are possible if you configure it that way.
        If that is the case, you will need to provide other identifying information such as O(private_ip_address) or O(eni_id).
    required: false
    type: str
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.tags
  - amazon.aws.boto3
notes:
  - This module identifies and ENI based on either the O(eni_id), a combination of O(private_ip_address) and O(subnet_id),
    or a combination of O(instance_id) and O(device_index). Any of these options will let you specify a particular ENI.
  - Support for O(tags) and O(purge_tags) was added in release 1.3.0.
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.

# Create an ENI. As no security group is defined, ENI will be created in default security group
- amazon.aws.ec2_eni:
    private_ip_address: 172.31.0.20
    subnet_id: subnet-xxxxxxxx
    state: present

# Create an ENI and attach it to an instance
- amazon.aws.ec2_eni:
    instance_id: i-xxxxxxx
    device_index: 1
    private_ip_address: 172.31.0.20
    subnet_id: subnet-xxxxxxxx
    state: present

# Create an ENI with two secondary addresses
- amazon.aws.ec2_eni:
    subnet_id: subnet-xxxxxxxx
    state: present
    secondary_private_ip_address_count: 2

# Assign a secondary IP address to an existing ENI
# This will purge any existing IPs
- amazon.aws.ec2_eni:
    subnet_id: subnet-xxxxxxxx
    eni_id: eni-yyyyyyyy
    state: present
    secondary_private_ip_addresses:
      - 172.16.1.1

# Remove any secondary IP addresses from an existing ENI
- amazon.aws.ec2_eni:
    subnet_id: subnet-xxxxxxxx
    eni_id: eni-yyyyyyyy
    state: present
    secondary_private_ip_address_count: 0

# Destroy an ENI, detaching it from any instance if necessary
- amazon.aws.ec2_eni:
    eni_id: eni-xxxxxxx
    force_detach: true
    state: absent

# Update an ENI
- amazon.aws.ec2_eni:
    eni_id: eni-xxxxxxx
    description: "My new description"
    state: present

# Update an ENI using name and subnet_id
- amazon.aws.ec2_eni:
    name: eni-20
    subnet_id: subnet-xxxxxxx
    description: "My new description"
    state: present

# Update an ENI identifying it by private_ip_address and subnet_id
- amazon.aws.ec2_eni:
    subnet_id: subnet-xxxxxxx
    private_ip_address: 172.16.1.1
    description: "My new description"

# Detach an ENI from an instance
- amazon.aws.ec2_eni:
    eni_id: eni-xxxxxxx
    instance_id: None
    state: present

### Delete an interface on termination
# First create the interface
- amazon.aws.ec2_eni:
    instance_id: i-xxxxxxx
    device_index: 1
    private_ip_address: 172.31.0.20
    subnet_id: subnet-xxxxxxxx
    state: present
  register: eni

# Modify the interface to enable the delete_on_terminaton flag
- amazon.aws.ec2_eni:
    eni_id: "{{ eni.interface.id }}"
    delete_on_termination: true
"""


RETURN = r"""
interface:
  description: Network interface attributes.
  returned: when state != absent
  type: complex
  contains:
    attachment:
      description: The network interface attachment.
      type: dict
      sample: {
            "attach_time": "2024-04-25T20:57:20+00:00",
            "attachment_id": "eni-attach-0ddce58b341a1846f",
            "delete_on_termination": true,
            "device_index": 0,
            "instance_id": "i-032cb1cceb29250d2",
            "status": "attached"
        }
    description:
      description: Interface description.
      type: str
      sample: Firewall network interface
    groups:
      description: Dict of security groups.
      type: dict
      sample: { "sg-f8a8a9da": "default" }
    id:
      description: Network interface id.
      type: str
      sample: "eni-1d889198"
    mac_address:
      description: Interface's physical address.
      type: str
      sample: "00:00:5E:00:53:23"
    name:
      description: The name of the ENI.
      type: str
      sample: "my-eni-20"
    owner_id:
      description: AWS account id.
      type: str
      sample: 812381371
    private_ip_address:
      description: Primary ip address of this interface.
      type: str
      sample: 10.20.30.40
    private_ip_addresses:
      description: List of all private ip addresses associated to this interface.
      type: list
      elements: dict
      sample: [ { "primary_address": true, "private_ip_address": "10.20.30.40" } ]
    source_dest_check:
      description: Value of source/dest check flag.
      type: bool
      sample: True
    status:
      description: Network interface status.
      type: str
      sample: "pending"
    subnet_id:
      description: Which vpc subnet the interface is bound.
      type: str
      sample: subnet-b0a0393c
    tags:
      description: The dictionary of tags associated with the ENI.
      type: dict
      sample: { "Name": "my-eni", "group": "Finance" }
    vpc_id:
      description: Which vpc this network interface is bound.
      type: str
      sample: vpc-9a9a9da
"""

import time
from ipaddress import ip_address
from ipaddress import ip_network

from ansible_collections.amazon.aws.plugins.module_utils.ec2 import AnsibleEC2Error
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import assign_private_ip_addresses
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import attach_network_interface
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import create_network_interface
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import delete_network_interface
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import describe_network_interfaces
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import describe_subnets
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import detach_network_interface
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import ensure_ec2_tags
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import get_ec2_security_group_ids_from_names
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import modify_network_interface_attribute
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import unassign_private_ip_addresses
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_specifications
from ansible_collections.amazon.aws.plugins.module_utils.waiters import wait_for_resource_state


def get_eni_info(interface):
    # Private addresses
    private_addresses = []
    if "PrivateIpAddresses" in interface:
        for ip in interface["PrivateIpAddresses"]:
            private_addresses.append({"private_ip_address": ip["PrivateIpAddress"], "primary_address": ip["Primary"]})

    groups = {}
    if "Groups" in interface:
        for group in interface["Groups"]:
            groups[group["GroupId"]] = group["GroupName"]

    interface_info = {
        "id": interface.get("NetworkInterfaceId"),
        "subnet_id": interface.get("SubnetId"),
        "vpc_id": interface.get("VpcId"),
        "description": interface.get("Description"),
        "owner_id": interface.get("OwnerId"),
        "status": interface.get("Status"),
        "mac_address": interface.get("MacAddress"),
        "private_ip_address": interface.get("PrivateIpAddress"),
        "source_dest_check": interface.get("SourceDestCheck"),
        "groups": groups,
        "private_ip_addresses": private_addresses,
    }

    if "TagSet" in interface:
        tags = boto3_tag_list_to_ansible_dict(interface["TagSet"])
        if "Name" in tags:
            interface_info["name"] = tags["Name"]
        interface_info["tags"] = tags

    if "Attachment" in interface:
        interface_info["attachment"] = {
            "attachment_id": interface["Attachment"].get("AttachmentId"),
            "instance_id": interface["Attachment"].get("InstanceId"),
            "device_index": interface["Attachment"].get("DeviceIndex"),
            "status": interface["Attachment"].get("Status"),
            "attach_time": interface["Attachment"].get("AttachTime"),
            "delete_on_termination": interface["Attachment"].get("DeleteOnTermination"),
        }

    return interface_info


def correct_ips(connection, ip_list, module, eni_id):
    eni = describe_eni(connection, module, eni_id)
    private_addresses = set()
    if "PrivateIpAddresses" in eni:
        for ip in eni["PrivateIpAddresses"]:
            private_addresses.add(ip["PrivateIpAddress"])

    ip_set = set(ip_list)

    return ip_set.issubset(private_addresses)


def absent_ips(connection, ip_list, module, eni_id):
    eni = describe_eni(connection, module, eni_id)
    private_addresses = set()
    if "PrivateIpAddresses" in eni:
        for ip in eni["PrivateIpAddresses"]:
            private_addresses.add(ip["PrivateIpAddress"])

    ip_set = set(ip_list)

    return not ip_set.union(private_addresses)


def correct_ip_count(connection, ip_count, module, eni_id):
    eni = describe_eni(connection, module, eni_id)
    private_addresses = set()
    if "PrivateIpAddresses" in eni:
        for ip in eni["PrivateIpAddresses"]:
            private_addresses.add(ip["PrivateIpAddress"])

    return bool(len(private_addresses) == ip_count)


def wait_for(function_pointer, *args):
    max_wait = 30
    interval_time = 3
    current_wait = 0
    while current_wait < max_wait:
        time.sleep(interval_time)
        current_wait += interval_time
        if function_pointer(*args):
            break


def create_eni(connection, vpc_id, module):
    instance_id = module.params.get("instance_id")
    attached = module.params.get("attached")
    if instance_id == "None":
        instance_id = None
    device_index = module.params.get("device_index")
    subnet_id = module.params.get("subnet_id")
    private_ip_address = module.params.get("private_ip_address")
    description = module.params.get("description")
    security_groups = get_ec2_security_group_ids_from_names(
        module.params.get("security_groups"), connection, vpc_id=vpc_id
    )
    secondary_private_ip_addresses = module.params.get("secondary_private_ip_addresses")
    secondary_private_ip_address_count = module.params.get("secondary_private_ip_address_count")
    changed = False

    tags = module.params.get("tags") or dict()
    name = module.params.get("name")
    # Make sure that the 'name' parameter sets the Name tag
    if name:
        tags["Name"] = name

    args = {"SubnetId": subnet_id}
    if private_ip_address:
        args["PrivateIpAddress"] = private_ip_address
    if description:
        args["Description"] = description
    if len(security_groups) > 0:
        args["Groups"] = security_groups
    if tags:
        args["TagSpecifications"] = boto3_tag_specifications(tags, types="network-interface")

    # check if provided private_ip_address is within the subnet's address range
    if private_ip_address:
        try:
            cidr_block = describe_subnets(connection, SubnetIds=[str(subnet_id)])[0]["CidrBlock"]
        except AnsibleEC2Error as e:
            module.fail_json_aws_error(e)
        valid_private_ip = ip_address(private_ip_address) in ip_network(cidr_block)
        if not valid_private_ip:
            module.fail_json(
                changed=False,
                msg="Error: cannot create ENI - Address does not fall within the subnet's address range.",
            )
    if module.check_mode:
        module.exit_json(changed=True, msg="Would have created ENI if not in check mode.")

    try:
        eni = create_network_interface(connection, **args)
    except AnsibleEC2Error as e:
        module.fail_json_aws_error(e)
    # Once we have an ID make sure we're always modifying the same object
    eni_id = eni["NetworkInterfaceId"]
    wait_for_resource_state(connection, module, "network_interface_available", NetworkInterfaceIds=[eni_id])

    if attached and instance_id is not None:
        try:
            attach_network_interface(
                connection,
                InstanceId=instance_id,
                DeviceIndex=device_index,
                NetworkInterfaceId=eni["NetworkInterfaceId"],
            )
        except AnsibleEC2Error as e:
            delete_network_interface(connection, eni_id)
            module.fail_json_aws_error(e)
        wait_for_resource_state(connection, module, "network_interface_attached", NetworkInterfaceIds=[eni_id])

    if secondary_private_ip_address_count is not None:
        try:
            assign_private_ip_addresses(
                connection,
                NetworkInterfaceId=eni["NetworkInterfaceId"],
                SecondaryPrivateIpAddressCount=secondary_private_ip_address_count,
            )
            wait_for(correct_ip_count, connection, secondary_private_ip_address_count, module, eni_id)
        except AnsibleEC2Error as e:
            delete_network_interface(connection, eni_id)
            module.fail_json_aws_error(e)

    if secondary_private_ip_addresses is not None:
        try:
            assign_private_ip_addresses(
                connection,
                NetworkInterfaceId=eni["NetworkInterfaceId"],
                PrivateIpAddresses=secondary_private_ip_addresses,
            )
            wait_for(correct_ips, connection, secondary_private_ip_addresses, module, eni_id)
        except AnsibleEC2Error as e:
            delete_network_interface(connection, eni_id)
            module.fail_json_aws_error(e)

    # Refresh the eni data
    eni = describe_eni(connection, module, eni_id)
    changed = True

    module.exit_json(changed=changed, interface=get_eni_info(eni))


def modify_eni(connection, module, eni):
    instance_id = module.params.get("instance_id")
    attached = module.params.get("attached")
    device_index = module.params.get("device_index")
    description = module.params.get("description")
    security_groups = module.params.get("security_groups")
    source_dest_check = module.params.get("source_dest_check")
    delete_on_termination = module.params.get("delete_on_termination")
    secondary_private_ip_addresses = module.params.get("secondary_private_ip_addresses")
    purge_secondary_private_ip_addresses = module.params.get("purge_secondary_private_ip_addresses")
    secondary_private_ip_address_count = module.params.get("secondary_private_ip_address_count")
    allow_reassignment = module.params.get("allow_reassignment")
    changed = False
    tags = module.params.get("tags")
    name = module.params.get("name")
    purge_tags = module.params.get("purge_tags")

    eni = uniquely_find_eni(connection, module, eni)
    eni_id = eni["NetworkInterfaceId"]

    try:
        if description is not None:
            if "Description" not in eni or eni["Description"] != description:
                if not module.check_mode:
                    modify_network_interface_attribute(
                        connection, NetworkInterfaceId=eni_id, Description={"Value": description}
                    )
                changed = True
        if len(security_groups) > 0:
            groups = get_ec2_security_group_ids_from_names(security_groups, connection, vpc_id=eni["VpcId"])
            if sorted(get_sec_group_list(eni["Groups"])) != sorted(groups):
                if not module.check_mode:
                    modify_network_interface_attribute(connection, NetworkInterfaceId=eni_id, Groups=groups)
                changed = True
        if source_dest_check is not None:
            if "SourceDestCheck" not in eni or eni["SourceDestCheck"] != source_dest_check:
                if not module.check_mode:
                    modify_network_interface_attribute(
                        connection, NetworkInterfaceId=eni_id, SourceDestCheck={"Value": source_dest_check}
                    )
                changed = True
        if delete_on_termination is not None and "Attachment" in eni:
            if eni["Attachment"]["DeleteOnTermination"] is not delete_on_termination:
                if not module.check_mode:
                    modify_network_interface_attribute(
                        connection,
                        NetworkInterfaceId=eni_id,
                        Attachment={
                            "AttachmentId": eni["Attachment"]["AttachmentId"],
                            "DeleteOnTermination": delete_on_termination,
                        },
                    )
                    if delete_on_termination:
                        waiter = "network_interface_delete_on_terminate"
                    else:
                        waiter = "network_interface_no_delete_on_terminate"
                    wait_for_resource_state(connection, module, waiter, NetworkInterfaceIds=[eni_id])
                changed = True

        current_secondary_addresses = []
        if "PrivateIpAddresses" in eni:
            current_secondary_addresses = [i["PrivateIpAddress"] for i in eni["PrivateIpAddresses"] if not i["Primary"]]

        if secondary_private_ip_addresses is not None:
            secondary_addresses_to_remove = list(set(current_secondary_addresses) - set(secondary_private_ip_addresses))
            if secondary_addresses_to_remove and purge_secondary_private_ip_addresses:
                if not module.check_mode:
                    unassign_private_ip_addresses(
                        connection,
                        NetworkInterfaceId=eni_id,
                        PrivateIpAddresses=list(set(current_secondary_addresses) - set(secondary_private_ip_addresses)),
                    )
                    wait_for(absent_ips, connection, secondary_addresses_to_remove, module, eni_id)
                changed = True
            secondary_addresses_to_add = list(set(secondary_private_ip_addresses) - set(current_secondary_addresses))
            if secondary_addresses_to_add:
                if not module.check_mode:
                    assign_private_ip_addresses(
                        connection,
                        NetworkInterfaceId=eni_id,
                        PrivateIpAddresses=secondary_addresses_to_add,
                        AllowReassignment=allow_reassignment,
                    )
                    wait_for(correct_ips, connection, secondary_addresses_to_add, module, eni_id)
                changed = True

        if secondary_private_ip_address_count is not None:
            current_secondary_address_count = len(current_secondary_addresses)
            if secondary_private_ip_address_count > current_secondary_address_count:
                if not module.check_mode:
                    assign_private_ip_addresses(
                        connection,
                        NetworkInterfaceId=eni_id,
                        SecondaryPrivateIpAddressCount=(
                            secondary_private_ip_address_count - current_secondary_address_count
                        ),
                        AllowReassignment=allow_reassignment,
                    )
                    wait_for(correct_ip_count, connection, secondary_private_ip_address_count, module, eni_id)
                changed = True
            elif secondary_private_ip_address_count < current_secondary_address_count:
                # How many of these addresses do we want to remove
                if not module.check_mode:
                    secondary_addresses_to_remove_count = (
                        current_secondary_address_count - secondary_private_ip_address_count
                    )
                    unassign_private_ip_addresses(
                        connection,
                        NetworkInterfaceId=eni_id,
                        PrivateIpAddresses=current_secondary_addresses[:secondary_addresses_to_remove_count],
                    )
                    wait_for(correct_ip_count, connection, secondary_private_ip_address_count, module, eni_id)
                changed = True

        if attached is True:
            if "Attachment" in eni and eni["Attachment"]["InstanceId"] != instance_id:
                if not module.check_mode:
                    detach_eni(connection, eni, module)
                    attach_network_interface(
                        connection,
                        InstanceId=instance_id,
                        DeviceIndex=device_index,
                        NetworkInterfaceId=eni_id,
                    )
                    wait_for_resource_state(
                        connection, module, "network_interface_attached", NetworkInterfaceIds=[eni_id]
                    )
                changed = True
            if "Attachment" not in eni:
                if not module.check_mode:
                    attach_network_interface(
                        connection,
                        InstanceId=instance_id,
                        DeviceIndex=device_index,
                        NetworkInterfaceId=eni_id,
                    )
                    wait_for_resource_state(
                        connection, module, "network_interface_attached", NetworkInterfaceIds=[eni_id]
                    )
                changed = True

        elif attached is False:
            changed |= detach_eni(connection, eni, module)
            wait_for_resource_state(connection, module, "network_interface_available", NetworkInterfaceIds=[eni_id])

        changed |= manage_tags(connection, module, eni, name, tags, purge_tags)

    except AnsibleEC2Error as e:
        module.fail_json_aws_error(e)

    eni = describe_eni(connection, module, eni_id)
    if module.check_mode and changed:
        module.exit_json(
            changed=changed, msg=f"Would have modified ENI: {eni['NetworkInterfaceId']} if not in check mode"
        )
    module.exit_json(changed=changed, interface=get_eni_info(eni))


def delete_eni(connection, module):
    eni = uniquely_find_eni(connection, module)
    if not eni:
        module.exit_json(changed=False)

    if module.check_mode:
        module.exit_json(changed=True, msg="Would have deleted ENI if not in check mode.")

    eni_id = eni["NetworkInterfaceId"]
    force_detach = module.params.get("force_detach")

    try:
        if force_detach is True:
            if "Attachment" in eni:
                detach_network_interface(
                    connection,
                    attachment_id=eni["Attachment"]["AttachmentId"],
                    force=True,
                )
                wait_for_resource_state(
                    connection,
                    module,
                    "network_interface_available",
                    delay=5,
                    max_attempts=80,
                    NetworkInterfaceIds=[eni_id],
                )
            changed = delete_network_interface(connection, eni_id)
        else:
            changed = delete_network_interface(connection, eni_id)

        module.exit_json(changed=changed)
    except AnsibleEC2Error as e:
        module.fail_json_aws_error(e)


def detach_eni(connection, eni, module):
    if module.check_mode:
        module.exit_json(changed=True, msg="Would have detached ENI if not in check mode.")

    eni_id = eni["NetworkInterfaceId"]

    force_detach = module.params.get("force_detach")
    if "Attachment" in eni:
        detach_network_interface(
            connection,
            attachment_id=eni["Attachment"]["AttachmentId"],
            force=force_detach,
        )
        wait_for_resource_state(
            connection, module, "network_interface_available", delay=5, max_attempts=80, NetworkInterfaceIds=[eni_id]
        )
        return True

    return False


def describe_eni(connection, module, eni_id):
    try:
        result = None
        enis = describe_network_interfaces(connection, NetworkInterfaceIds=[eni_id])
        if enis:
            result = enis[0]
        return result
    except AnsibleEC2Error as e:
        module.fail_json_aws_error(e)


def uniquely_find_eni(connection, module, eni=None):
    if eni:
        # In the case of create, eni_id will not be a param but we can still get the eni_id after creation
        if "NetworkInterfaceId" in eni:
            eni_id = eni["NetworkInterfaceId"]
        else:
            eni_id = None
    else:
        eni_id = module.params.get("eni_id")

    private_ip_address = module.params.get("private_ip_address")
    subnet_id = module.params.get("subnet_id")
    instance_id = module.params.get("instance_id")
    device_index = module.params.get("device_index")
    attached = module.params.get("attached")
    name = module.params.get("name")

    filters = []

    # proceed only if we're unequivocally specifying an ENI
    if eni_id is None and private_ip_address is None and (instance_id is None and device_index is None):
        return None

    if eni_id:
        filters.append({"Name": "network-interface-id", "Values": [eni_id]})

    if private_ip_address and subnet_id and not filters:
        filters.append({"Name": "private-ip-address", "Values": [private_ip_address]})
        filters.append({"Name": "subnet-id", "Values": [subnet_id]})

    if not attached and instance_id and device_index and not filters:
        filters.append({"Name": "attachment.instance-id", "Values": [instance_id]})
        filters.append({"Name": "attachment.device-index", "Values": [str(device_index)]})

    if name and subnet_id and not filters:
        filters.append({"Name": "tag:Name", "Values": [name]})
        filters.append({"Name": "subnet-id", "Values": [subnet_id]})

    if not filters:
        return None

    try:
        result = None
        enis = describe_network_interfaces(connection, Filters=filters)
        if len(enis) == 1:
            result = enis[0]
        return result
    except AnsibleEC2Error as e:
        module.fail_json_aws_error(e)


def get_sec_group_list(groups):
    # Build list of remote security groups
    remote_security_groups = []
    for group in groups:
        remote_security_groups.append(group["GroupId"])

    return remote_security_groups


def _get_vpc_id(connection, module, subnet_id):
    try:
        return describe_subnets(connection, SubnetIds=[subnet_id])[0]["VpcId"]
    except AnsibleEC2Error as e:
        module.fail_json_aws_error(e)


def manage_tags(connection, module, eni, name, tags, purge_tags):
    # Do not purge tags unless tags is not None
    if tags is None:
        purge_tags = False
        tags = {}

    if name:
        tags["Name"] = name

    eni_id = eni["NetworkInterfaceId"]

    changed = ensure_ec2_tags(connection, module, eni_id, tags=tags, purge_tags=purge_tags)
    return changed


def main():
    argument_spec = dict(
        eni_id=dict(default=None, type="str"),
        instance_id=dict(default=None, type="str"),
        private_ip_address=dict(type="str"),
        subnet_id=dict(type="str"),
        description=dict(type="str"),
        security_groups=dict(default=[], type="list", elements="str"),
        device_index=dict(default=0, type="int"),
        state=dict(default="present", choices=["present", "absent"]),
        force_detach=dict(default="no", type="bool"),
        source_dest_check=dict(default=None, type="bool"),
        delete_on_termination=dict(default=None, type="bool"),
        secondary_private_ip_addresses=dict(default=None, type="list", elements="str"),
        purge_secondary_private_ip_addresses=dict(default=False, type="bool"),
        secondary_private_ip_address_count=dict(default=None, type="int"),
        allow_reassignment=dict(default=False, type="bool"),
        attached=dict(default=None, type="bool"),
        name=dict(default=None, type="str"),
        tags=dict(type="dict", aliases=["resource_tags"]),
        purge_tags=dict(default=True, type="bool"),
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        mutually_exclusive=[["secondary_private_ip_addresses", "secondary_private_ip_address_count"]],
        required_if=(
            [
                ("attached", True, ["instance_id"]),
                ("purge_secondary_private_ip_addresses", True, ["secondary_private_ip_addresses"]),
            ]
        ),
        supports_check_mode=True,
    )

    connection = module.client("ec2")
    state = module.params.get("state")

    if state == "present":
        eni = uniquely_find_eni(connection, module)
        if eni is None:
            subnet_id = module.params.get("subnet_id")
            if subnet_id is None:
                module.fail_json(msg="subnet_id is required when creating a new ENI")

            vpc_id = _get_vpc_id(connection, module, subnet_id)
            create_eni(connection, vpc_id, module)
        else:
            modify_eni(connection, module, eni)

    elif state == "absent":
        delete_eni(connection, module)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright: (c) 2024, Samita Bhattacharjee (@samitab) <samitab.cisco.com>

# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "certified"}

DOCUMENTATION = r"""
---
module: aci_fabric_pod_connectivity_profile
short_description: Manage Fabric External Pod Connectivity Profiles (fv:PodConnP)
description:
- Manage Fabric External Pod Connectivity Profiles on Cisco ACI fabrics.
options:
  pod_id:
    description:
    - The Pod ID associated with the Pod Connectivity Profile.
    type: int
    aliases: [ pod, pid ]
  fabric_id:
    description:
    - The Fabric ID associated with the Pod Connectivity Profile.
    type: int
    aliases: [ fabric, fid ]
  virtual_pod_id:
    description:
    - The Pod ID in the main fabric to which this I(pod_id) is associated. This property is valid only if this pod is a virtual pod.
    type: int
    aliases: [ vpod, vpod_id ]
  description:
    description:
    - The description of the Pod Connectivity Profile.
    type: str
    aliases: [ descr ]
  data_plane_tep:
    description:
    - The Data Plane TEP IPv4 address and prefix.
    - eg. 10.1.1.1/32
    type: str
    aliases: [ dp_tep ]
  unicast_tep:
    description:
    - The Unicast TEP IPv4 address and prefix.
    - eg. 10.1.1.2/32
    type: str
    aliases: [ u_tep ]
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation
- cisco.aci.owner

notes:
- This module requires an existing I(fabric_external_connection_profile).
  The module M(cisco.aci.aci_fabric_external_connection_profile) can be used for this.
seealso:
- module: cisco.aci.aci_fabric_external_connection_profile
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(fv:PodConnP).
  link: https://developer.cisco.com/docs/apic-mim-ref/
author:
- Samita Bhattacharjee (@samitab)
"""

EXAMPLES = r"""
- name: Add a Pod Connectivity Profile
  cisco.aci.aci_fabric_pod_connectivity_profile:
    host: apic
    username: admin
    password: SomeSecretPassword
    fabric_id: 1
    pod_id: 1
    description: First pod connectivity profile
    data_plane_tep: 10.1.1.1/32
    unicast_tep: 10.1.1.2/32
    state: present
  delegate_to: localhost

- name: Query a Pod Connectivity Profile
  cisco.aci.aci_fabric_pod_connectivity_profile:
    host: apic
    username: admin
    password: SomeSecretPassword
    fabric_id: 1
    pod_id: 1
    state: query
  delegate_to: localhost
  register: query_result

- name: Query all Pod Connectivity Profiles
  cisco.aci.aci_fabric_pod_connectivity_profile:
    host: apic
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost
  register: query_result

- name: Remove a Pod Connectivity Profile
  cisco.aci.aci_fabric_pod_connectivity_profile:
    host: apic
    username: admin
    password: SomeSecretPassword
    fabric_id: 1
    pod_id: 1
    state: absent
  delegate_to: localhost
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec, aci_owner_spec


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(aci_owner_spec())
    argument_spec.update(
        pod_id=dict(type="int", aliases=["pod", "pid"]),
        fabric_id=dict(type="int", aliases=["fabric", "fid"]),
        virtual_pod_id=dict(type="int", aliases=["vpod", "vpod_id"]),
        description=dict(type="str", aliases=["descr"]),
        data_plane_tep=dict(type="str", aliases=["dp_tep"]),
        unicast_tep=dict(type="str", aliases=["u_tep"]),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["fabric_id", "pod_id"]],
            ["state", "present", ["fabric_id", "pod_id"]],
        ],
    )

    aci = ACIModule(module)

    pod_id = module.params.get("pod_id")
    fabric_id = module.params.get("fabric_id")
    virtual_pod_id = module.params.get("virtual_pod_id")
    description = module.params.get("description")
    data_plane_tep = module.params.get("data_plane_tep")
    unicast_tep = module.params.get("unicast_tep")
    state = module.params.get("state")

    aci.construct_url(
        root_class=dict(
            aci_class="fvFabricExtConnP",
            aci_rn="tn-infra/fabricExtConnP-{0}".format(fabric_id),
            module_object=fabric_id,
            target_filter={"id": fabric_id},
        ),
        subclass_1=dict(
            aci_class="fvPodConnP",
            aci_rn="podConnP-{0}".format(pod_id),
            module_object=pod_id,
            target_filter={"id": pod_id},
        ),
        child_classes=["fvIp", "fvExtRoutableUcastConnP"],
    )

    aci.get_existing()

    if state == "present":
        child_configs = []

        # Validate if existing and remove child objects when the config does not match the provided config.
        if isinstance(aci.existing, list) and len(aci.existing) > 0:
            for child in aci.existing[0].get("fvPodConnP", {}).get("children", {}):
                if child.get("fvExtRoutableUcastConnP") and child.get("fvExtRoutableUcastConnP").get("attributes").get("addr") != unicast_tep:
                    child_configs.append(
                        {
                            "fvExtRoutableUcastConnP": {
                                "attributes": {
                                    "addr": child.get("fvExtRoutableUcastConnP").get("attributes").get("addr"),
                                    "status": "deleted",
                                }
                            }
                        }
                    )
                if child.get("fvIp") and child.get("fvIp").get("attributes").get("addr") != data_plane_tep:
                    child_configs.append(
                        {
                            "fvIp": {
                                "attributes": {
                                    "addr": child.get("fvIp").get("attributes").get("addr"),
                                    "status": "deleted",
                                }
                            }
                        }
                    )

        if unicast_tep is not None:
            child_configs.append({"fvExtRoutableUcastConnP": {"attributes": {"addr": unicast_tep}}})
        if data_plane_tep is not None:
            child_configs.append({"fvIp": {"attributes": {"addr": data_plane_tep}}})

        aci.payload(
            aci_class="fvPodConnP",
            class_config=dict(
                id=pod_id,
                assocIntersitePodId=virtual_pod_id,
                descr=description,
            ),
            child_configs=child_configs,
        )

        aci.get_diff(aci_class="fvPodConnP")

        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



# Copyright (c) 2020 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible.errors import AnsibleFilterError
from ansible.module_utils.common.collections import is_sequence
from ansible.module_utils.common.text.converters import to_text

from ..plugin_utils._quote import (
    quote_c,
    quote_cmd,
    quote_pwsh,
)


def quote(value, shell=None):
    """Quotes argument(s) for the various shells in Windows command processing.

    Quotes argument(s) for the various Windows command line shells. Default to escaping arguments based on the Win32 C
    argv parsing rules that 'win_command' uses but shell='cmd' or shell='powershell' can be set to escape arguments for
    those respective shells. Each value is escaped in a way to ensure the process gets the literal argument passed in
    and meta chars escaped.

    When passing in a dict, the arguments will be in the form 'key={{ value | ansible.windows.quote }}' to match the
    MSI parameter format.

    :params value: A string, list, or dict of value(s) to quote.
    :params shell: The shell that is used to escape the args for.
    :return: The quoted argument(s) from the input.
    """
    if not shell:
        quote_func = quote_c
    elif shell == 'cmd':
        quote_func = quote_cmd
    elif shell == 'powershell':
        quote_func = quote_pwsh
    else:
        raise AnsibleFilterError("Invalid shell specified, valid shell are None, 'cmd', or 'powershell'")

    if not is_sequence(value):
        value = [value]

    new_arguments = []
    for arg in value:
        if isinstance(arg, dict):
            # Quoting a dict assumes you are quoting a KEY=value pair for MSI arguments. The key can only be
            # '[A-Z0-9_\.]' so we don't attempt to quote the key, only the value. If another format is desired then it
            # need to be done manually, i.e. '/KEY:{{ value | ansible.windows.quote }}'.
            for key, value in arg.items():
                k = to_text(key, errors='surrogate_or_strict')
                v = quote_func(to_text(value, errors='surrogate_or_strict', nonstring='passthru'))
                new_arguments.append(u'%s=%s' % (k, v))
        else:
            new_arguments.append(quote_func(to_text(arg, errors='surrogate_or_strict', nonstring='passthru')))

    return u" ".join(new_arguments)


class FilterModule:

    def filters(self):
        return {
            'quote': quote,
        }



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)


DOCUMENTATION = r"""
---
module: issues_info
short_description: Information module for Issues Info
description:
- This module represents an alias of the module issues_v1_info
version_added: '3.1.0'
extends_documentation_fragment:
  - cisco.dnac.module_info
author: Rafael Campos (@racampos)
options:
  headers:
    description: Additional headers.
    type: dict
  startTime:
    description:
    - StartTime query parameter. Starting epoch time in milliseconds of query time window.
    type: float
  endTime:
    description:
    - EndTime query parameter. Ending epoch time in milliseconds of query time window.
    type: float
  siteId:
    description:
    - SiteId query parameter. Assurance UUID value of the site in the issue content.
    type: str
  deviceId:
    description:
    - DeviceId query parameter. Assurance UUID value of the device in the issue content.
    type: str
  macAddress:
    description:
    - MacAddress query parameter. Client's device MAC address of the issue (format xx xx xx xx xx xx).
    type: str
  priority:
    description:
    - >
      Priority query parameter. The issue's priority value P1, P2, P3, or P4 (case insensitive) (Use only when
      macAddress and deviceId are not provided).
    type: str
  issueStatus:
    description:
    - IssueStatus query parameter. The issue's status value ACTIVE, IGNORED, RESOLVED (case insensitive).
    type: str
  aiDriven:
    description:
    - >
      AiDriven query parameter. The issue's AI driven value YES or NO (case insensitive) (Use only when macAddress
      and deviceId are not provided).
    type: str
requirements:
- dnacentersdk >= 2.4.9
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Issues IssuesV1
  description: Complete reference of the IssuesV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!issues
notes:
  - SDK Method used are
    issues.Issues.issues_v1,

  - Paths used are
    get /dna/intent/api/v1/issues,
  - It should be noted that this module is an alias of issues_v1_info

"""

EXAMPLES = r"""
- name: Get all Issues Info
  cisco.dnac.issues_info:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    headers: "{{my_headers | from_json}}"
    startTime: 0
    endTime: 0
    siteId: string
    deviceId: string
    macAddress: string
    priority: string
    issueStatus: string
    aiDriven: string
  register: result

"""
RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "version": "string",
      "totalCount": "string",
      "response": [
        {
          "issueId": "string",
          "name": "string",
          "siteId": "string",
          "deviceId": "string",
          "deviceRole": "string",
          "aiDriven": "string",
          "clientMac": "string",
          "issue_occurence_count": 0,
          "status": "string",
          "priority": "string",
          "category": "string",
          "last_occurence_time": 0
        }
      ]
    }
"""



# -*- coding: utf-8 -*-
# Copyright 2024 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The iosxr route_maps fact class
It is in this file the configuration is collected from the device
for a given resource, parsed, and the facts tree is populated
based on the configuration.
"""


from ansible_collections.ansible.netcommon.plugins.module_utils.network.common import utils

from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.argspec.route_maps.route_maps import (
    Route_mapsArgs,
)
from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.rm_templates.route_maps import (
    Route_mapsTemplate,
)


class Route_mapsFacts(object):
    """The iosxr route_maps facts class"""

    def __init__(self, module, subspec="config", options="options"):
        self._module = module
        self.argument_spec = Route_mapsArgs.argument_spec

    def get_policynames(self, connection):
        return connection.get("show running-config | include route-policy")

    def get_policydata(self, connection, name):
        return connection.get(f"show running-config route-policy {name}")

    def parse_condition(self, condition):
        if condition.startswith("if "):
            condition_type, cond = "if_section", (condition.lstrip("if ")).rstrip(" then")
        elif condition.startswith("elseif "):
            condition_type, cond = "elseif_section", (condition.lstrip("elseif ")).rstrip(" then")
        elif condition.startswith("else"):
            condition_type, cond = "else_section", ""
        elif condition.startswith("global"):
            condition_type, cond = "global", ""
        return condition_type, cond

    def parse_route_policy(self, route_policy):
        """This would take one route policy as input and process the configurations
        and group them by the conditions, the sections that don't have a condition
        are grouped into global and the configuration with condition are grouped into
        the condition iteslf. This does not invoke the parsers it just groups the
        condition with the config lines for the condition under route-policy


        :param route_policy: raw single route-policy configuration

        :rtype: dictionary
        :returns: processed route policy
        """
        result = {}
        lines = route_policy.splitlines()
        current_key = None
        current_value = []
        store_global = True
        global_value = []
        else_data = {}

        def process_else(else_line):
            """this is a separate implementation as else deals with a few parts of config differently"""
            else_result = {}
            else_current_key = None
            else_current_value = []
            else_store_global = True
            else_global_value = []

            for line in else_line:
                line = line.strip()
                if line.startswith("if ") or line.startswith("elseif ") or line.startswith("else"):
                    else_store_global = False
                    if else_current_key:
                        else_result[else_current_key] = else_current_value
                    else_current_key = line
                    else_current_value = []
                else:
                    else_current_value.append(line)

                if else_store_global:
                    else_global_value.append(line)

                if else_global_value:
                    else_result["global"] = else_global_value

            if else_current_key:
                else_result[else_current_key] = else_current_value

            return else_result

        for idx, line in enumerate(lines):
            line = line.strip()

            if line.startswith("if ") or line.startswith("elseif "):
                store_global = False
                if current_key:
                    result[current_key] = current_value
                current_key = line
                current_value = []
            else:
                current_value.append(line)

            if not (line.startswith("if ") or line.startswith("elseif ")) and line.startswith(
                "else",
            ):
                else_data = process_else(lines[idx + 1 : :])  # noqa: E203
                break

            if store_global:
                global_value.append(line)

        # Add the last block
        if current_key:
            result[current_key] = current_value

        if else_data:
            result["else"] = else_data

        if global_value:
            result["global"] = global_value

        return result

    def get_policy_config(self, policy_data, name):
        """Facts for individual policy is generated here, and extends the route_maps facts"""
        policy_map_structured = self.parse_route_policy(policy_data)

        def else_resolve_policy_data(else_policy_map):
            """Handles else segment, quite similar but different as else in else behaves differently"""
            else_policy_route = {}
            if_elif = []

            for condition, policy in else_policy_map.items():
                if_elif_data = {}
                cond_type, actual_cond = self.parse_condition(condition)

                route_maps_parser = Route_mapsTemplate(lines=policy, module=self._module)
                objs = list(route_maps_parser.parse().values())
                if cond_type in ["if_section", "global"]:
                    if objs:
                        else_policy_route[cond_type] = objs[0]
                        if cond_type != "global":
                            else_policy_route[cond_type]["condition"] = actual_cond
                elif cond_type == "elseif_section":
                    if_elif_data.update(objs[0])
                    if_elif_data["condition"] = actual_cond
                    if_elif.append(if_elif_data)
                elif cond_type == "else_section":
                    if objs:
                        else_policy_route[cond_type] = objs[0]

            if if_elif:
                else_policy_route["elseif_section"] = if_elif

            return else_policy_route

        def rec_resolve_policy_data(policy_map):
            """resolved each policy condition and data and parses policy configuration and sieves out condition data"""
            policy_route = {
                "name": name,
            }
            if_elif = []
            else_data = {}

            for condition, policy in policy_map.items():
                if_elif_data = {}
                cond_type, actual_cond = self.parse_condition(condition)

                route_maps_parser = Route_mapsTemplate(lines=policy, module=self._module)
                objs = list(route_maps_parser.parse().values())
                if cond_type in ["if_section", "global"]:
                    if objs:
                        policy_route[cond_type] = objs[0]
                        if cond_type != "global":
                            policy_route[cond_type]["condition"] = actual_cond
                elif cond_type == "elseif_section":
                    if_elif_data.update(objs[0])
                    if_elif_data["condition"] = actual_cond
                    if_elif.append(if_elif_data)
                elif cond_type == "else_section":
                    else_data = else_resolve_policy_data(policy)

            if if_elif:
                policy_route["elseif_section"] = if_elif
            if else_data:
                policy_route["else_section"] = else_data

            return policy_route

        rec_policy_route = rec_resolve_policy_data(policy_map_structured)

        return rec_policy_route

    def populate_facts(self, connection, ansible_facts, data=None):
        """Populate the facts for Route_maps network resource

        :param connection: the device connection
        :param ansible_facts: Facts dictionary
        :param data: previously collected conf

        :rtype: dictionary
        :returns: facts
        """
        facts = {}
        objs = []
        policy_names = []
        mock_data = False

        if not data:
            # gets policy names as there is no good way to get all policy data at once,
            # other than slicing running-config
            data = self.get_policynames(connection=connection)
        else:
            mock_data = True  # for states like parsed to work

        # parse native config using the Route_maps template
        route_maps_parser = Route_mapsTemplate(lines=[], module=self._module)

        for name in data.splitlines():  # generate a list of policy names
            if name.startswith("route-policy "):
                policy_names.append(name.split()[1])

        if mock_data:  # only for states like parsed
            data_for_parsed = data.split("end-policy\n!")

        for idx, policy in enumerate(policy_names):
            if mock_data:
                # we enumerate the split data as the name and policy details are on the same sequence
                policy_data = data_for_parsed[idx]
            else:
                # we send the name of the policy and the policy data is fetched for each name, one costly operation
                policy_data = self.get_policydata(connection=connection, name=policy)
            # the list of policy facts is created as individual route-policy information is converted to facts
            objs.append(self.get_policy_config(policy_data=policy_data, name=policy))

        ansible_facts["ansible_network_resources"].pop("route_maps", None)

        params = utils.remove_empties(
            route_maps_parser.validate_config(self.argument_spec, {"config": objs}, redact=True),
        )

        facts["route_maps"] = params.get("config", [])  # handles empty config
        ansible_facts["ansible_network_resources"].update(facts)

        return ansible_facts



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: sda_fabrics_vlan_to_ssids_fabric_id_v1_info
short_description: Information module for Sda Fabrics Vlan To Ssids Fabric Id V1
description:
- Get all Sda Fabrics Vlan To Ssids Fabric Id V1.
- >
   Retrieve the VLANs and SSIDs mapped to the VLAN, within a Fabric Site. The 'fabricId' represents the Fabric ID of
   a particular Fabric Site.
version_added: '6.15.0'
extends_documentation_fragment:
  - cisco.dnac.module_info
author: Rafael Campos (@racampos)
options:
  headers:
    description: Additional headers.
    type: dict
  fabricId:
    description:
    - FabricId path parameter. The 'fabricId' represents the Fabric ID of a particular Fabric Site.
    type: str
  limit:
    description:
    - Limit query parameter. The number of records to show for this page.
    type: float
  offset:
    description:
    - Offset query parameter. The first record to show for this page; the first record is numbered 1.
    type: float
requirements:
- dnacentersdk >= 2.4.9
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Fabric Wireless RetrieveTheVLANsAndSSIDsMappedToTheVLANWithinAFabricSiteV1
  description: Complete reference of the RetrieveTheVLANsAndSSIDsMappedToTheVLANWithinAFabricSiteV1 API.
  link: https://developer.cisco.com/docs/dna-center/#!retrieve-the-vla-ns-and-ssi-ds-mapped-to-the-vlan-within-a-fabric-site  # noqa: E501
notes:
  - SDK Method used are
    fabric_wireless.FabricWireless.retrieve_the_vlans_and_ssids_mapped_to_the_vlan_within_a_fabric_site_v1,

  - Paths used are
    get /dna/intent/api/v1/sda/fabrics/{fabricId}/vlanToSsids,

"""

EXAMPLES = r"""
- name: Get all Sda Fabrics Vlan To Ssids Fabric Id V1
  cisco.dnac.sda_fabrics_vlan_to_ssids_fabric_id_v1_info:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    headers: "{{my_headers | from_json}}"
    limit: 0
    offset: 0
    fabricId: string
  register: result

"""
RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": [
        {
          "vlanName": "string",
          "ssidDetails": [
            {
              "name": "string",
              "securityGroupTag": "string"
            }
          ]
        }
      ],
      "version": "string"
    }
"""



#!/usr/bin/python
#
# Copyright (c) 2019 Yuwei Zhou, <yuwzho@microsoft.com>
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_loganalyticsworkspace_info
version_added: "0.1.2"
short_description: Get facts of Azure Log Analytics workspaces
description:
    - Get, query Azure Log Analytics workspaces.
options:
    resource_group:
        description:
            - Name of resource group.
        required: True
        type: str
    name:
        description:
            - Name of the workspace.
        type: str
    tags:
        description:
            - Limit results by providing a list of tags. Format tags as 'key' or 'key:value'.
        type: list
        elements: str
    show_intelligence_packs:
        description:
            - Show the intelligence packs for a workspace.
            - Note this will cost one more network overhead for each workspace, expected slow response.
        type: bool
    show_management_groups:
        description:
            - Show the management groups for a workspace.
            - Note this will cost one more network overhead for each workspace, expected slow response.
        type: bool
    show_shared_keys:
        description:
            - Show the shared keys for a workspace.
            - Note this will cost one more network overhead for each workspace, expected slow response.
        type: bool
    show_usages:
        description:
            - Show the list of usages for a workspace.
            - Note this will cost one more network overhead for each workspace, expected slow response.
        type: bool
extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Yuwei Zhou (@yuwzho)

'''

EXAMPLES = '''
- name: Query a workspace
  azure_rm_loganalyticsworkspace_info:
      resource_group: myResourceGroup
      name: myLogAnalyticsWorkspace
      show_intelligence_packs: true
      show_management_groups: true
      show_shared_keys: true
      show_usages: true
'''

RETURN = '''
id:
    description:
        - Workspace resource path.
    type: str
    returned: success
    example: "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/myResourceGroup/providers/Microsoft.OperationalInsights/workspaces/m
              yLogAnalyticsWorkspace"
location:
    description:
        - Resource location.
    type: str
    returned: success
    example: "eastus"
sku:
    description:
        - The SKU of the workspace.
    type: str
    returned: success
    example: "per_gb2018"
retention_in_days:
    description:
        - The workspace data retention in days.
        - -1 means Unlimited retention for I(sku=unlimited).
        - 730 days is the maximum allowed for all other SKUs.
    type: int
    returned: success
    example: 40
intelligence_packs:
    description:
        - Lists all the intelligence packs possible and whether they are enabled or disabled for a given workspace.
    type: list
    returned: success
    example: [ {'name': 'CapacityPerformance', 'enabled': true} ]
management_groups:
    description:
        - Management groups connected to the workspace.
    type: dict
    returned: success
    example: {'value': []}
shared_keys:
    description:
        - Shared keys for the workspace.
    type: dict
    returned: success
    example: {
                'primarySharedKey': 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx',
                'secondarySharedKey': 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx'
              }
usages:
    description:
        - Usage metrics for the workspace.
    type: dict
    returned: success
    example: {
                'value': [
                    {
                    'name': {
                        'value': 'DataAnalyzed',
                        'localizedValue': 'Data Analyzed'
                    },
                    'unit': 'Bytes',
                    'currentValue': 0,
                    'limit': 524288000,
                    'nextResetTime': '2017-10-03T00:00:00Z',
                    'quotaPeriod': 'P1D'
                    }
                ]
              }
'''  # NOQA

from ansible.module_utils.common.dict_transformations import _camel_to_snake

try:
    from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase
    from azure.core.exceptions import ResourceNotFoundError
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMLogAnalyticsWorkspaceInfo(AzureRMModuleBase):

    def __init__(self):

        self.module_arg_spec = dict(
            resource_group=dict(type='str', required=True),
            name=dict(type='str'),
            tags=dict(type='list', elements='str'),
            show_shared_keys=dict(type='bool'),
            show_intelligence_packs=dict(type='bool'),
            show_usages=dict(type='bool'),
            show_management_groups=dict(type='bool')
        )

        self.results = dict(
            changed=False,
            workspaces=[]
        )

        self.resource_group = None
        self.name = None
        self.tags = None
        self.show_intelligence_packs = None
        self.show_shared_keys = None
        self.show_usages = None
        self.show_management_groups = None

        super(AzureRMLogAnalyticsWorkspaceInfo, self).__init__(self.module_arg_spec, supports_check_mode=True, supports_tags=False, facts_module=True)

    def exec_module(self, **kwargs):

        is_old_facts = self.module._name == 'azure_rm_loganalyticsworkspace_facts'
        if is_old_facts:
            self.module.deprecate("The 'azure_rm_loganalyticsworkspace_facts' module has been renamed to 'azure_rm_loganalyticsworkspace_info'",
                                  version=(2.9, ))

        for key in list(self.module_arg_spec.keys()):
            setattr(self, key, kwargs[key])

        if self.name:
            item = self.get_workspace()
            response = [item] if item else []
        else:
            response = self.list_by_resource_group()

        self.results['workspaces'] = [self.to_dict(x) for x in response if self.has_tags(x.tags, self.tags)]
        return self.results

    def get_workspace(self):
        try:
            return self.log_analytics_client.workspaces.get(self.resource_group, self.name)
        except ResourceNotFoundError:
            pass
        return None

    def list_by_resource_group(self):
        try:
            return self.log_analytics_client.resource_group.list(self.resource_group)
        except Exception:
            pass
        return []

    def list_intelligence_packs(self):
        try:
            response = self.log_analytics_client.intelligence_packs.list(self.resource_group, self.name)
            return [x.as_dict() for x in response]
        except Exception as exc:
            self.fail('Error when listing intelligence packs {0}'.format(exc.message or str(exc)))

    def list_management_groups(self):
        result = []
        try:
            response = self.log_analytics_client.management_groups.list(self.resource_group, self.name)
            while True:
                result.append(response.next().as_dict())
        except StopIteration:
            pass
        except Exception as exc:
            self.fail('Error when listing management groups {0}'.format(exc.message or str(exc)))
        return result

    def list_usages(self):
        result = []
        try:
            response = self.log_analytics_client.usages.list(self.resource_group, self.name)
            while True:
                result.append(response.next().as_dict())
        except StopIteration:
            pass
        except Exception as exc:
            self.fail('Error when listing usages {0}'.format(exc.message or str(exc)))
        return result

    def get_shared_keys(self):
        try:
            return self.log_analytics_client.shared_keys.get_shared_keys(self.resource_group, self.name).as_dict()
        except Exception as exc:
            self.fail('Error when getting shared key {0}'.format(exc.message or str(exc)))

    def to_dict(self, workspace):
        result = workspace.as_dict()
        result['sku'] = _camel_to_snake(workspace.sku.name)
        if self.show_intelligence_packs:
            result['intelligence_packs'] = self.list_intelligence_packs()
        if self.show_management_groups:
            result['management_groups'] = self.list_management_groups()
        if self.show_shared_keys:
            result['shared_keys'] = self.get_shared_keys()
        if self.show_usages:
            result['usages'] = self.list_usages()
        return result


def main():
    AzureRMLogAnalyticsWorkspaceInfo()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# coding: utf-8 -*-

# (c) 2017, Wayne Witzel III <wayne@riotousliving.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1', 'status': ['preview'], 'supported_by': 'community'}


DOCUMENTATION = '''
---
module: job_wait
author: "Wayne Witzel III (@wwitzel3)"
short_description: Wait for Automation Platform Controller job to finish.
description:
    - Wait for Automation Platform Controller job to finish and report success or failure. See
      U(https://www.ansible.com/tower) for an overview.
options:
    job_id:
      description:
        - ID of the job to monitor.
      required: True
      type: int
    interval:
      description:
        - The interval in sections, to request an update from the controller.
        - For backwards compatibility if unset this will be set to the average of min and max intervals
      required: False
      default: 2
      type: float
    timeout:
      description:
        - Maximum time in seconds to wait for a job to finish.
      type: int
    job_type:
      description:
        - Job type to wait for
      choices: ['project_updates', 'jobs', 'inventory_updates', 'workflow_jobs']
      default: 'jobs'
      type: str
extends_documentation_fragment: awx.awx.auth
'''

EXAMPLES = '''
- name: Launch a job
  job_launch:
    job_template: "My Job Template"
  register: job

- name: Wait for job max 120s
  job_wait:
    job_id: "{{ job.id }}"
    timeout: 120
'''

RETURN = '''
id:
    description: job id that is being waited on
    returned: success
    type: int
    sample: 99
elapsed:
    description: total time in seconds the job took to run
    returned: success
    type: float
    sample: 10.879
started:
    description: timestamp of when the job started running
    returned: success
    type: str
    sample: "2017-03-01T17:03:53.200234Z"
finished:
    description: timestamp of when the job finished running
    returned: success
    type: str
    sample: "2017-03-01T17:04:04.078782Z"
status:
    description: current status of job
    returned: success
    type: str
    sample: successful
'''


from ..module_utils.controller_api import ControllerAPIModule


def main():
    # Any additional arguments that are not fields of the item can be added here
    argument_spec = dict(
        job_id=dict(type='int', required=True),
        job_type=dict(choices=['project_updates', 'jobs', 'inventory_updates', 'workflow_jobs'], default='jobs'),
        timeout=dict(type='int'),
        interval=dict(type='float', default=2),
    )

    # Create a module for ourselves
    module = ControllerAPIModule(argument_spec=argument_spec)

    # Extract our parameters
    job_id = module.params.get('job_id')
    job_type = module.params.get('job_type')
    timeout = module.params.get('timeout')
    interval = module.params.get('interval')

    # Attempt to look up job based on the provided id
    job = module.get_one(
        job_type,
        **{
            'data': {
                'id': job_id,
            }
        }
    )

    if job is None:
        module.fail_json(msg='Unable to wait on ' + job_type.rstrip("s") + ' {0}; that ID does not exist.'.format(job_id))

    # Invoke wait function
    module.wait_on_url(url=job['url'], object_name=job_id, object_type='legacy_job_wait', timeout=timeout, interval=interval)

    module.exit_json(**module.json_output)


if __name__ == '__main__':
    main()



# Copyright (c) 2018 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = r'''
    name: azure_rm
    short_description: Azure Resource Manager inventory plugin
    extends_documentation_fragment:
      - azure.azcollection.azure
      - azure.azcollection.azure_rm
      - constructed
    description:
        - Query VM details from Azure Resource Manager
        - Requires a YAML configuration file whose name ends with 'azure_rm.(yml|yaml)'
        - By default, sets C(ansible_host) to the first public IP address found (preferring the primary NIC). If no
          public IPs are found, the first private IP (also preferring the primary NIC). The default may be overridden
          via C(hostvar_expressions); see examples.
'''

EXAMPLES = '''
# The following host variables are always available:
# public_ipv4_addresses: all public IP addresses, with the primary IP config from the primary NIC first
# public_dns_hostnames: all public DNS hostnames, with the primary IP config from the primary NIC first
# private_ipv4_addresses: all private IP addressses, with the primary IP config from the primary NIC first
# id: the VM's Azure resource ID, eg /subscriptions/00000000-0000-0000-1111-1111aaaabb/resourceGroups/my_rg/providers/Microsoft.Compute/virtualMachines/my_vm
# location: the VM's Azure location, eg 'westus', 'eastus'
# name: the VM's resource name, eg 'myvm'
# os_profile: The VM OS properties, a dictionary, only system is currently available, eg 'os_profile.system not in ['linux']'
# powerstate: the VM's current power state, eg: 'running', 'stopped', 'deallocated'
# provisioning_state: the VM's current provisioning state, eg: 'succeeded'
# tags: dictionary of the VM's defined tag values
# resource_type: the VM's resource type, eg: 'Microsoft.Compute/virtualMachine', 'Microsoft.Compute/virtualMachineScaleSets/virtualMachines',
# 'microsoft.azurestackhci/virtualmachineinstances'
# vmid: the VM's internal SMBIOS ID, eg: '36bca69d-c365-4584-8c06-a62f4a1dc5d2'
# vmss: if the VM is a member of a scaleset (vmss), a dictionary including the id and name of the parent scaleset
# availability_zone: availability zone in which VM is deployed, eg '1','2','3'
# creation_time: datetime object of when the VM was created, eg '2023-07-21T09:30:30.4710164+00:00'
#
# The following host variables are sometimes available:
# computer_name: the Operating System's hostname. Will not be available if azure agent is not available and picking it up.
# The following host variables are available for Azure Stack HCI vms:
# customLocation: the azure arc custom location.
# virtual_machine_memoryMB: RAM allowed (static)
# virtual_machine_processors: number of vCPUs


# sample 'myazuresub.azure_rm.yaml'

# required for all azure_rm inventory plugin configs
plugin: azure.azcollection.azure_rm

# forces this plugin to use a CLI auth session instead of the automatic auth source selection (eg, prevents the
# presence of 'ANSIBLE_AZURE_RM_X' environment variables from overriding CLI auth)
auth_source: cli

# fetches VMs from an explicit list of resource groups instead of default all (- '*')
include_vm_resource_groups:
    - myrg1
    - myrg2

# fetches VMs from VMSSs in all resource groups (defaults to no VMSS fetch)
include_vmss_resource_groups:
    - '*'

# fetches VMs from Azure StackHCI in specific resource groups (defaults to no HCI vm fetch)
include_hcivm_resource_groups:
    - myrg1

# places a host in the named group if the associated condition evaluates to true
conditional_groups:
    # since this will be true for every host, every host sourced from this inventory plugin config will be in the
    # group 'all_the_hosts'
    all_the_hosts: true
    # if the VM's "name" variable contains "dbserver", it will be placed in the 'db_hosts' group
    db_hosts: "'dbserver' in name"

# adds variables to each host found by this inventory plugin, whose values are the result of the associated expression
hostvar_expressions:
    my_host_var:
    # A statically-valued expression has to be both single and double-quoted, or use escaped quotes, since the outer
    # layer of quotes will be consumed by YAML. Without the second set of quotes, it interprets 'staticvalue' as a
    # variable instead of a string literal.
    some_statically_valued_var: "'staticvalue'"
    # overrides the default ansible_host value with a custom Jinja2 expression, in this case, the first DNS hostname, or
    # if none are found, the first public IP address.
    ansible_host: (public_dns_hostnames + public_ipv4_addresses) | first

# change how inventory_hostname is generated. Each item is a jinja2 expression similar to hostvar_expressions.
hostnames:
    - tags.vm_name
    - default_inventory_hostname + ".domain.tld" # Transfer to fqdn if you use shortnames for VMs
    - default  # special var that uses the default hashed name

# places hosts in dynamically-created groups based on a variable value.
keyed_groups:
# places each host in a group named 'tag_(tag name)_(tag value)' for each tag on a VM.
    - prefix: tag
      key: tags
# places each host in a group named 'azure_loc_(location name)', depending on the VM's location
    - prefix: azure_loc
      key: location
# places host in a group named 'some_tag_X' using the value of the 'sometag' tag on a VM as X, and defaulting to the
# value 'none' (eg, the group 'some_tag_none') if the 'sometag' tag is not defined for a VM.
    - prefix: some_tag
      key: tags.sometag | default('none')

# excludes a host from the inventory when any of these expressions is true, can refer to any vars defined on the host
exclude_host_filters:
    # excludes hosts in the eastus region
    - location in ['eastus']
    - tags['tagkey'] is defined and tags['tagkey'] == 'tagvalue'
    - tags['tagkey2'] is defined and tags['tagkey2'] == 'tagvalue2'
    # excludes hosts that are powered off
    - powerstate != 'running'

# includes a host to the inventory when any of these expressions is true, can refer to any vars defined on the host
include_host_filters:
    # includes hosts that in the eastus region and power on
    - location in ['eastus'] and powerstate == 'running'
    # includes hosts in the eastus region and power on OR includes hosts in the eastus2 region and tagkey value is tagvalue
    - location in ['eastus'] and powerstate == 'running'
    - location in ['eastus2'] and tags['tagkey'] is defined and tags['tagkey'] == 'tagvalue'
'''

# FUTURE: do we need a set of sane default filters, separate from the user-defineable ones?
# eg, powerstate==running, provisioning_state==succeeded


import hashlib
import json
import re
import uuid
import os
import time

try:
    from queue import Queue, Empty
except ImportError:
    from Queue import Queue, Empty

from collections import namedtuple
from ansible.plugins.inventory import BaseInventoryPlugin, Constructable
from ansible.module_utils.six import iteritems
from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMAuth
from ansible.errors import AnsibleParserError, AnsibleError
from ansible.module_utils.parsing.convert_bool import boolean
from ansible.module_utils._text import to_native, to_bytes, to_text
from itertools import chain
from os import environ

try:
    from azure.core._pipeline_client import PipelineClient
    from azure.core.pipeline.policies import BearerTokenCredentialPolicy
    from azure.core.configuration import Configuration
    from azure.mgmt.core.tools import parse_resource_id
    from azure.core.pipeline import PipelineResponse
    from azure.mgmt.core.polling.arm_polling import ARMPolling
    from azure.core.polling import LROPoller
except ImportError:
    Configuration = object
    parse_resource_id = object
    PipelineClient = object
    BearerTokenCredentialPolicy = object
    pass


class AzureRMRestConfiguration(Configuration):
    def __init__(self, credentials, subscription_id, base_url=None):

        if credentials is None:
            raise ValueError("Parameter 'credentials' must not be None.")
        if subscription_id is None:
            raise ValueError("Parameter 'subscription_id' must not be None.")
        if not base_url:
            base_url = 'https://management.azure.com'

        credential_scopes = base_url + '/.default'

        super(AzureRMRestConfiguration, self).__init__()

        self.authentication_policy = BearerTokenCredentialPolicy(credentials, credential_scopes)
        self.credentials = credentials
        self.subscription_id = subscription_id


UrlAction = namedtuple('UrlAction', ['url', 'api_version', 'handler', 'handler_args'])


# FUTURE: add Cacheable support once we have a sane serialization format
class InventoryModule(BaseInventoryPlugin, Constructable):

    NAME = 'azure.azcollection.azure_rm'

    def __init__(self):
        super(InventoryModule, self).__init__()

        self._hosts = []
        self._filters = None

        # FUTURE: use API profiles with defaults
        self._compute_api_version = '2021-11-01'
        self._network_api_version = '2015-06-15'
        self._hybridcompute_api_version = '2024-05-20-preview'
        self._stackhci_api_version = '2024-01-01'

        self._default_header_parameters = {'Content-Type': 'application/json; charset=utf-8'}

        self._request_queue = Queue()

        self.azure_auth = None

        self._batch_fetch = False

    def verify_file(self, path):
        '''
            :param loader: an ansible.parsing.dataloader.DataLoader object
            :param path: the path to the inventory config file
            :return the contents of the config file
        '''
        if super(InventoryModule, self).verify_file(path):
            if re.match(r'.{0,}azure_rm\.y(a)?ml$', path):
                return True
        # display.debug("azure_rm inventory filename must end with 'azure_rm.yml' or 'azure_rm.yaml'")
        raise AnsibleError("azure_rm inventory filename must end with 'azure_rm.yml' or 'azure_rm.yaml'")

    def parse(self, inventory, loader, path, cache=True):
        super(InventoryModule, self).parse(inventory, loader, path)

        self._read_config_data(path)

        if self.get_option('use_contrib_script_compatible_sanitization'):
            self._sanitize_group_name = self._legacy_script_compatible_group_sanitization

        self._batch_fetch = self.get_option('batch_fetch')

        self._legacy_hostnames = self.get_option('plain_host_names')

        self._filters = self.get_option('exclude_host_filters') + self.get_option('default_host_filters')

        self._include_filters = self.get_option('include_host_filters')

        try:
            self._credential_setup()
            self._get_hosts()
        except Exception:
            raise

    def _credential_setup(self):
        auth_source = environ.get('ANSIBLE_AZURE_AUTH_SOURCE', None) or self.get_option('auth_source')
        auth_options = dict(
            auth_source=auth_source,
            profile=self.get_option('profile'),
            subscription_id=self.get_option('subscription_id'),
            client_id=self.get_option('client_id'),
            secret=self.get_option('secret'),
            tenant=self.get_option('tenant'),
            ad_user=self.get_option('ad_user'),
            password=self.get_option('password'),
            cloud_environment=self.get_option('cloud_environment'),
            cert_validation_mode=self.get_option('cert_validation_mode'),
            api_profile=self.get_option('api_profile'),
            track1_cred=True,
            adfs_authority_url=self.get_option('adfs_authority_url')
        )

        if self.templar.is_template(auth_options["tenant"]):
            auth_options["tenant"] = self.templar.template(variable=auth_options["tenant"], disable_lookups=False)

        if self.templar.is_template(auth_options["client_id"]):
            auth_options["client_id"] = self.templar.template(variable=auth_options["client_id"], disable_lookups=False)

        if self.templar.is_template(auth_options["secret"]):
            auth_options["secret"] = self.templar.template(variable=auth_options["secret"], disable_lookups=False)

        if self.templar.is_template(auth_options["subscription_id"]):
            auth_options["subscription_id"] = self.templar.template(variable=auth_options["subscription_id"], disable_lookups=False)

        self.azure_auth = AzureRMAuth(**auth_options)

        self._clientconfig = AzureRMRestConfiguration(self.azure_auth.azure_credential_track2, self.azure_auth.subscription_id,
                                                      self.azure_auth._cloud_environment.endpoints.resource_manager)

        self.new_client = PipelineClient(self.azure_auth._cloud_environment.endpoints.resource_manager, config=self._clientconfig)

    def _enqueue_get(self, url, api_version, handler, handler_args=None):
        if not handler_args:
            handler_args = {}
        self._request_queue.put_nowait(UrlAction(url=url, api_version=api_version, handler=handler, handler_args=handler_args))

    def _enqueue_vm_list(self, rg='*'):
        if not rg or rg == '*':
            url = '/subscriptions/{subscriptionId}/providers/Microsoft.Compute/virtualMachines'
        else:
            url = '/subscriptions/{subscriptionId}/resourceGroups/{rg}/providers/Microsoft.Compute/virtualMachines'

        url = url.format(subscriptionId=self._clientconfig.subscription_id, rg=rg)
        self._enqueue_get(url=url, api_version=self._compute_api_version, handler=self._on_vm_page_response)

    def _enqueue_arcvm_list(self, rg='*'):
        if not rg or rg == '*':
            url = '/subscriptions/{subscriptionId}/providers/Microsoft.HybridCompute/machines'
        else:
            url = '/subscriptions/{subscriptionId}/resourceGroups/{rg}/providers/Microsoft.HybridCompute/machines'

        url = url.format(subscriptionId=self._clientconfig.subscription_id, rg=rg)
        self._enqueue_get(url=url, api_version=self._hybridcompute_api_version, handler=self._on_arcvm_page_response)

    def _enqueue_vmss_list(self, rg=None):
        if not rg or rg == '*':
            url = '/subscriptions/{subscriptionId}/providers/Microsoft.Compute/virtualMachineScaleSets'
        else:
            url = '/subscriptions/{subscriptionId}/resourceGroups/{rg}/providers/Microsoft.Compute/virtualMachineScaleSets'

        url = url.format(subscriptionId=self._clientconfig.subscription_id, rg=rg)
        self._enqueue_get(url=url, api_version=self._compute_api_version, handler=self._on_vmss_page_response)

    def _get_hosts(self):
        if os.environ.get('ANSIBLE_AZURE_VM_RESOURCE_GROUPS'):
            for vm_rg in os.environ['ANSIBLE_AZURE_VM_RESOURCE_GROUPS'].split(","):
                self._enqueue_vm_list(vm_rg)
        else:
            for vm_rg in self.get_option('include_vm_resource_groups'):
                self._enqueue_vm_list(vm_rg)

        for vm_rg in self.get_option('include_hcivm_resource_groups'):
            self._enqueue_arcvm_list(vm_rg)

        if os.environ.get('ANSIBLE_AZURE_VMSS_RESOURCE_GROUPS'):
            for vmss_rg in os.environ['ANSIBLE_AZURE_VMSS_RESOURCE_GROUPS'].split(","):
                self._enqueue_vmss_list(vmss_rg)
        else:
            for vmss_rg in self.get_option('include_vmss_resource_groups'):
                self._enqueue_vmss_list(vmss_rg)

        if self._batch_fetch:
            self._process_queue_batch()
        else:
            self._process_queue_serial()

        constructable_config_strict = boolean(self.get_option('fail_on_template_errors'))
        if self.get_option('hostvar_expressions') is not None:
            constructable_config_compose = self.get_option('hostvar_expressions')
        else:
            constructable_config_compose = self.get_option('compose')
        constructable_config_groups = self.get_option('conditional_groups')
        constructable_config_keyed_groups = self.get_option('keyed_groups')

        constructable_hostnames = self.get_option('hostnames')

        for h in self._hosts:
            # FUTURE: track hostnames to warn if a hostname is repeated (can happen for legacy and for composed inventory_hostname)
            inventory_hostname = self._get_hostname(h, hostnames=constructable_hostnames, strict=constructable_config_strict)
            if self._filter_exclude_host(inventory_hostname, h.hostvars):
                continue
            if not self._filter_include_host(inventory_hostname, h.hostvars):
                continue
            self.inventory.add_host(inventory_hostname)
            # FUTURE: configurable default IP list? can already do this via hostvar_expressions
            self.inventory.set_variable(inventory_hostname, "ansible_host",
                                        next(chain(h.hostvars['public_ipv4_address'], h.hostvars['private_ipv4_addresses']), None))
            for k, v in iteritems(h.hostvars):
                # FUTURE: configurable hostvar prefix? Makes docs harder...
                self.inventory.set_variable(inventory_hostname, k, v)

            # constructable delegation
            self._set_composite_vars(constructable_config_compose, h.hostvars, inventory_hostname, strict=constructable_config_strict)
            self._add_host_to_composed_groups(constructable_config_groups, h.hostvars, inventory_hostname, strict=constructable_config_strict)
            self._add_host_to_keyed_groups(constructable_config_keyed_groups, h.hostvars, inventory_hostname, strict=constructable_config_strict)

    # FUTURE: fix underlying inventory stuff to allow us to quickly access known groupvars from reconciled host
    def _filter_host(self, filter, inventory_hostname, hostvars):
        self.templar.available_variables = hostvars

        for condition in filter:
            # FUTURE: should warn/fail if conditional doesn't return True or False
            conditional = "{{% if {0} %}} True {{% else %}} False {{% endif %}}".format(condition)
            try:
                if boolean(self.templar.template(conditional)):
                    return True
            except Exception as e:
                if boolean(self.get_option('fail_on_template_errors')):
                    raise AnsibleParserError("Error evaluating filter condition '{0}' for host {1}: {2}".format(condition, inventory_hostname, to_native(e)))
                continue

        return False

    def _filter_include_host(self, inventory_hostname, hostvars):
        return self._filter_host(self._include_filters, inventory_hostname, hostvars)

    def _filter_exclude_host(self, inventory_hostname, hostvars):
        return self._filter_host(self._filters, inventory_hostname, hostvars)

    def _get_hostname(self, host, hostnames=None, strict=False):
        hostname = None
        errors = []

        for preference in hostnames:
            if preference == 'default':
                return host.default_inventory_hostname
            try:
                hostname = self._compose(preference, host.hostvars)
            except Exception as e:  # pylint: disable=broad-except
                if strict:
                    raise AnsibleError("Could not compose %s as hostnames - %s" % (preference, to_native(e)))
                else:
                    errors.append(
                        (preference, str(e))
                    )
            if hostname:
                return to_text(hostname)

        raise AnsibleError(
            'Could not template any hostname for host, errors for each preference: %s' % (
                ', '.join(['%s: %s' % (pref, err) for pref, err in errors])
            )
        )

    def _process_queue_serial(self):
        try:
            while True:
                item = self._request_queue.get_nowait()
                resp = self.send_request(item.url, item.api_version)
                item.handler(resp, **item.handler_args)
        except Empty:
            pass

    def _on_vm_page_response(self, response, vmss=None, arcvm=None):
        next_link = response.get('nextLink')

        if next_link:
            self._enqueue_get(url=next_link, api_version=self._compute_api_version, handler=self._on_vm_page_response,
                              handler_args=dict(vmss=vmss, arcvm=arcvm))

        if 'value' in response:
            for h in response['value']:
                # FUTURE: add direct VM filtering by tag here (performance optimization)?
                self._hosts.append(AzureHost(h, self, vmss=vmss, arcvm=arcvm, legacy_name=self._legacy_hostnames))

    def _on_arcvm_page_response(self, response):
        next_link = response.get('nextLink')

        if next_link:
            self._enqueue_get(url=next_link, api_version=self._hybridcompute_api_version, handler=self._on_arcvm_page_response)

        for arcvm in response['value']:
            url = '{0}/providers/Microsoft.AzureStackHCI/virtualMachineInstances'.format(arcvm['id'])
            # Stack HCI instances look close enough to regular VMs that we can share the handler impl...
            self._enqueue_get(url=url, api_version=self._stackhci_api_version, handler=self._on_vm_page_response, handler_args=dict(arcvm=arcvm))

    def _on_vmss_page_response(self, response):
        next_link = response.get('nextLink')

        if next_link:
            self._enqueue_get(url=next_link, api_version=self._compute_api_version, handler=self._on_vmss_page_response)

        # FUTURE: add direct VMSS filtering by tag here (performance optimization)?
        for vmss in response['value']:
            url = '{0}/virtualMachines'.format(vmss['id'])

            # Since Flexible instance is a standalone VM we are processing them as regular VM.
            if vmss['properties']['orchestrationMode'] != 'Flexible':
                # VMSS instances look close enough to regular VMs that we can share the handler impl...
                self._enqueue_get(url=url, api_version=self._compute_api_version, handler=self._on_vm_page_response, handler_args=dict(vmss=vmss))

    # use the undocumented /batch endpoint to bulk-send up to 500 requests in a single round-trip
    #
    def _process_queue_batch(self):
        while True:
            batch_requests = []
            batch_item_index = 0
            batch_response_handlers = dict()
            try:
                while batch_item_index < 100:
                    item = self._request_queue.get_nowait()

                    name = str(uuid.uuid4())
                    query_parameters = {'api-version': item.api_version}
                    header_parameters = {'x-ms-client-request-id': str(uuid.uuid4()), 'Content-Type': 'application/json; charset=utf-8'}
                    body = {}
                    req = self.new_client.get(item.url, query_parameters, header_parameters, body)
                    batch_requests.append(dict(httpMethod="GET", url=req.url, name=name))
                    batch_response_handlers[name] = item
                    batch_item_index += 1
            except Empty:
                pass

            if not batch_requests:
                break

            self.retry_batch(batch_requests, batch_response_handlers)

    def retry_batch(self, batch_requests, batch_response_handlers, backoff_factor=0.8, retry_limit=10):
        retry_count = 1
        _SAFE_CODES = set(range(506)) - set([408, 429, 500, 502, 503, 504])
        _RETRY_CODES = set(range(999)) - _SAFE_CODES
        while True:
            batch_resp = self._send_batch(batch_requests)
            key_name = None
            if 'responses' in batch_resp:
                key_name = 'responses'
            elif 'value' in batch_resp:
                key_name = 'value'
            else:
                raise AnsibleError("didn't find expected key responses/value in batch response")
            batch_processed = []
            batch_retry = False
            for idx, r in enumerate(batch_resp[key_name]):
                status_code = r.get('httpStatusCode')
                returned_name = r['name']
                result = batch_response_handlers[returned_name]
                if status_code == 200:
                    # FUTURE: error-tolerant operation mode (eg, permissions)
                    # FUTURE: store/handle errors from individual handlers
                    result.handler(r['content'], **result.handler_args)
                    batch_processed.append(returned_name)
                elif status_code in _RETRY_CODES:
                    # 429: Too many requests Error, Backoff and Retry
                    batch_retry = True
            if batch_retry:
                time.sleep(backoff_factor * (2 ** (retry_count)))
                retry_count += 1
                if len(batch_processed) > 0:
                    # Remove already processed requests
                    for idx, r in enumerate(batch_requests):
                        if r.get('name') in batch_processed:
                            processed = batch_requests.pop(idx)
                if retry_count > retry_limit:
                    raise AnsibleError("Reached maximum retries in batch request")
            else:
                break

    def _send_batch(self, batched_requests):
        url = '/batch'
        query_parameters = {'api-version': '2015-11-01'}
        header_parameters = {'x-ms-client-request-id': str(uuid.uuid4()), 'Content-Type': 'application/json; charset=utf-8'}
        polling_timeout = 600
        polling_interval = 30
        operation_config = {}
        body_content = dict(requests=batched_requests)

        header = {'x-ms-client-request-id': str(uuid.uuid4())}
        header.update(self._default_header_parameters)

        request_new = self.new_client.post(url, query_parameters, header_parameters, body_content)
        response = self.new_client.send_request(request_new, **operation_config)

        if response.status_code == 202:
            def get_long_running_output(response):
                return response
            poller = LROPoller(self.new_client,
                               PipelineResponse(None, response, None),
                               get_long_running_output,
                               ARMPolling(polling_interval, **operation_config))
            response = self.get_poller_result(poller, polling_timeout)
            if hasattr(response, 'body'):
                response = json.loads(response.body())
            elif hasattr(response, 'context'):
                response = response.context['deserialized_data']
        else:
            response = json.loads(response.body())

        return response

    def get_poller_result(self, poller, timeout):
        try:
            poller.wait(timeout=timeout)
            return poller.result()
        except Exception as exc:
            raise

    def send_request(self, url, api_version):
        query_parameters = {'api-version': api_version}
        header_parameters = {'x-ms-client-request-id': str(uuid.uuid4()), 'Content-Type': 'application/json; charset=utf-8'}
        body = {}
        request_new = self.new_client.get(url, query_parameters, header_parameters)
        response = self.new_client.send_request(request_new)

        return json.loads(response.body())

    @staticmethod
    def _legacy_script_compatible_group_sanitization(name):

        # note that while this mirrors what the script used to do, it has many issues with unicode and usability in python
        regex = re.compile(r"[^A-Za-z0-9\_\-]")

        return regex.sub('_', name)

# VM list (all, N resource groups): VM -> InstanceView, N NICs, N PublicIPAddress)
# VMSS VMs (all SS, N specific SS, N resource groups?): SS -> VM -> InstanceView, N NICs, N PublicIPAddress)


class AzureHost(object):
    _powerstate_regex = re.compile('^PowerState/(?P<powerstate>.+)$')

    def __init__(self, vm_model, inventory_client, vmss=None, arcvm=None, legacy_name=False):
        self._inventory_client = inventory_client
        self._vm_model = vm_model
        self._vmss = vmss
        self._arcvm = arcvm

        self._instanceview = None

        self._powerstate = "unknown"
        self.nics = []

        vm_name = self._arcvm['name'] if self._arcvm else self._vm_model['name']

        if legacy_name:
            self.default_inventory_hostname = vm_name
        else:
            # Azure often doesn't provide a globally-unique filename, so use resource name + a chunk of ID hash
            self.default_inventory_hostname = '{0}_{1}'.format(vm_name, hashlib.sha1(to_bytes(vm_model['id'])).hexdigest()[0:4])

        self._hostvars = {}

        if self._arcvm:
            self._instanceview = self._vm_model
            self._powerstate = self._vm_model['properties'].get('status', {}).get('powerState', '').lower()  # 'Running'
        else:
            inventory_client._enqueue_get(url="{0}/instanceView".format(vm_model['id']),
                                          api_version=self._inventory_client._compute_api_version,
                                          handler=self._on_instanceview_response)

        nic_refs = vm_model['properties']['networkProfile']['networkInterfaces']
        for nic in nic_refs:
            # single-nic instances don't set primary, so figure it out...
            is_primary = nic.get('properties', {}).get('primary', len(nic_refs) == 1)
            api_version = self._inventory_client._stackhci_api_version if self._arcvm else self._inventory_client._network_api_version
            inventory_client._enqueue_get(url=nic['id'],
                                          api_version=api_version,
                                          handler=self._on_nic_response,
                                          handler_args=dict(is_primary=is_primary))

    @property
    def hostvars(self):
        if self._hostvars != {}:
            return self._hostvars

        system = "unknown"
        if self._arcvm and self._arcvm['properties'].get('osType'):  # osType unavailable with disabled guest agent
            system = self._arcvm['properties']['osType']
        elif 'osProfile' in self._vm_model['properties']:
            if 'linuxConfiguration' in self._vm_model['properties']['osProfile']:
                system = 'linux'
            if 'windowsConfiguration' in self._vm_model['properties']['osProfile']:
                system = 'windows'
        else:
            osType = self._vm_model['properties']['storageProfile']['osDisk']['osType']
            if osType == 'Linux':
                system = 'linux'
            if osType == 'Windows':
                system = 'windows'
        av_zone = None
        if 'zones' in self._vm_model:
            av_zone = self._vm_model['zones']

        createdAt = self._vm_model.get('systemData', {}).get('createdAt')  # hci specific

        new_hostvars = dict(
            network_interface=[],
            mac_address=[],
            network_interface_id=[],
            security_group_id=[],
            security_group=[],
            public_ip_address=[],
            public_ipv4_address=[],
            public_dns_hostnames=[],
            private_ipv4_addresses=[],
            subnet=[],
            id=self._vm_model['id'],
            location=self._arcvm['location'] if self._arcvm else self._vm_model['location'],
            name=self._arcvm['name'] if self._arcvm else self._vm_model['name'],
            computer_name=self._vm_model['properties'].get('osProfile', {}).get('computerName'),
            availability_zone=av_zone,
            powerstate=self._powerstate,
            provisioning_state=self._vm_model['properties']['provisioningState'].lower(),
            tags=self._arcvm.get('tags', {}) if self._arcvm else self._vm_model.get('tags', {}),
            resource_type=self._vm_model.get('type', "unknown"),
            vmid=self._vm_model['properties']['vmId'],
            os_profile=dict(
                system=system,
            ),
            vmss=dict(
                id=self._vmss['id'],
                name=self._vmss['name'],
            ) if self._vmss else {},
            virtual_machine_size=self._vm_model['properties']['hardwareProfile']['vmSize'] if self._vm_model['properties'].get('hardwareProfile') else None,
            plan=self._vm_model['properties']['plan']['name'] if self._vm_model['properties'].get('plan') else None,
            resource_group=parse_resource_id(self._vm_model['id']).get('resource_group').lower(),
            default_inventory_hostname=self.default_inventory_hostname,
            creation_time=createdAt if createdAt else self._vm_model['properties'].get('timeCreated'),
            license_type=self._vm_model['properties'].get('licenseType', 'Unknown')
        )
        if self._arcvm:
            new_hostvars['customLocation'] = self._vm_model.get('extendedLocation', {}).get('name', '').split('/')[-1]
            new_hostvars['virtual_machine_memoryMB'] = self._vm_model['properties']['hardwareProfile'].get('memoryMB')
            new_hostvars['virtual_machine_processors'] = self._vm_model['properties']['hardwareProfile'].get('processors')

        # set nic-related values from the primary NIC first
        for nic in sorted(self.nics, key=lambda n: n.is_primary, reverse=True):
            # and from the primary IP config per NIC first
            for ipc in sorted(nic._nic_model.get('properties', {}).get('ipConfigurations', []),
                              key=lambda i: i.get('properties', {}).get('primary', False), reverse=True):
                try:
                    subnet = ipc['properties'].get('subnet')
                    if subnet:
                        new_hostvars['subnet'].append(subnet)
                    private_ip = ipc['properties'].get('privateIPAddress')
                    if private_ip:
                        new_hostvars['private_ipv4_addresses'].append(private_ip)
                    pip_id = ipc['properties'].get('publicIPAddress', {}).get('id')
                    if pip_id and pip_id in nic.public_ips:
                        pip = nic.public_ips[pip_id]
                        new_hostvars['public_ipv4_address'].append(pip._pip_model['properties'].get('ipAddress', None))
                        new_hostvars['public_ip_address'].append({
                            'id': pip_id,
                            'name': pip._pip_model['name'],
                            'ipv4_address': pip._pip_model['properties'].get('ipAddress', None),
                        })
                        pip_fqdn = pip._pip_model['properties'].get('dnsSettings', {}).get('fqdn')
                        if pip_fqdn:
                            new_hostvars['public_dns_hostnames'].append(pip_fqdn)
                except Exception:
                    continue

            new_hostvars['mac_address'].append(nic._nic_model['properties'].get('macAddress'))
            new_hostvars['network_interface'].append(nic._nic_model['name'])
            new_hostvars['network_interface_id'].append(nic._nic_model['id'])
            new_hostvars['security_group_id'].append(nic._nic_model['properties']['networkSecurityGroup']['id']) \
                if nic._nic_model['properties'].get('networkSecurityGroup') else None
            new_hostvars['security_group'].append(parse_resource_id(nic._nic_model['properties']['networkSecurityGroup']['id'])['resource_name']) \
                if nic._nic_model['properties'].get('networkSecurityGroup') else None

        # set image and os_disk
        new_hostvars['image'] = {}
        new_hostvars['os_disk'] = {}
        new_hostvars['data_disks'] = []
        storageProfile = self._vm_model['properties'].get('storageProfile')
        if storageProfile:
            imageReference = storageProfile.get('imageReference')
            if imageReference:
                if imageReference.get('publisher'):
                    new_hostvars['image'] = dict(
                        sku=imageReference.get('sku'),
                        publisher=imageReference.get('publisher'),
                        version=imageReference.get('version'),
                        offer=imageReference.get('offer')
                    )
                elif imageReference.get('id'):
                    new_hostvars['image'] = dict(
                        id=imageReference.get('id')
                    )

            osDisk = storageProfile.get('osDisk')
            new_hostvars['os_disk'] = dict(
                name=osDisk.get('name'),
                operating_system_type=osDisk.get('osType').lower() if osDisk.get('osType') else None,
                id=storageProfile.get('vmConfigStoragePathId') if self._arcvm else osDisk.get('managedDisk', {}).get('id')
            )

            if self._arcvm:
                new_hostvars['data_disks'] = [
                    dict(
                        name=dataDisk.get('id').split('/')[-1],
                        id=dataDisk.get('id')
                    ) for dataDisk in storageProfile.get('dataDisks', [])
                ]
            else:
                new_hostvars['data_disks'] = [
                    dict(
                        name=dataDisk.get('name'),
                        lun=dataDisk.get('lun'),
                        id=dataDisk.get('managedDisk', {}).get('id')
                    ) for dataDisk in storageProfile.get('dataDisks', [])
                ]

        self._hostvars = new_hostvars
        return self._hostvars

    def _on_instanceview_response(self, vm_instanceview_model):
        self._instanceview = vm_instanceview_model
        self._powerstate = next((self._powerstate_regex.match(s.get('code', '')).group('powerstate')
                                 for s in vm_instanceview_model.get('statuses', []) if self._powerstate_regex.match(s.get('code', ''))), 'unknown')

    def _on_nic_response(self, nic_model, is_primary=False):
        nic = AzureNic(nic_model=nic_model, inventory_client=self._inventory_client, is_primary=is_primary)
        self.nics.append(nic)


class AzureNic(object):
    def __init__(self, nic_model, inventory_client, is_primary=False):
        self._nic_model = nic_model
        self.is_primary = is_primary
        self._inventory_client = inventory_client

        self.public_ips = {}

        if nic_model.get('properties', {}).get('ipConfigurations'):
            for ipc in nic_model['properties']['ipConfigurations']:
                pip = ipc['properties'].get('publicIPAddress')
                if pip:
                    self._inventory_client._enqueue_get(url=pip['id'], api_version=self._inventory_client._network_api_version, handler=self._on_pip_response)

    def _on_pip_response(self, pip_model):
        self.public_ips[pip_model['id']] = AzurePip(pip_model)


class AzurePip(object):
    def __init__(self, pip_model):
        self._pip_model = pip_model
